# Copyright (C) 2012-2013  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem
from lmi_client_base import LmiBaseClient
from lmi_client_base import LmiReturnValue
from lmi_address import LmiIpv4Addr

class LmiIpClient(LmiBaseClient):
    LMI_ETHERNET_PORT_CLASS_NAME = "LMI_EthernetPort"
    LMI_IP_ASSIGNMENT_SETTING_DATA_CLASS_NAME = "LMI_IPAssignmentSettingData"
    IP_CONFIGURATION_SERVICE_CLASS_NAME = "LMI_IPConfigurationService"
    IP_PROTOCOL_ENDPOINT_CLASS_NAME = "LMI_IPProtocolEndpoint"
    IP_PROTOCOL_TYPE_IPV4 = 1
    IP_PROTOCOL_TYPE_IPV6 = 2

    def __init__(self, hostname, username = "", password = ""):
        super(self.__class__, self).__init__(hostname, username, password)

    def __ip_get_ip_protocol_endpoint_instances(self, device = ""):
        (inst_list, rparams, errorstr) = self._get_instances(LmiIpClient.IP_PROTOCOL_ENDPOINT_CLASS_NAME)
        if not inst_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = errorstr)
        rval_inst_list = filter(lambda n: device == n["Name"].split("_")[0], inst_list) if device else inst_list
        if not rval_inst_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = "No such device '%s' found" % device)
        return LmiReturnValue(rval = rval_inst_list if len(rval_inst_list) > 1 else rval_inst_list[0],
            rparams = rparams, errorstr = "")

    def __ip_print_device_info(self, device):
        device_list = device if type(device) == list else [device]
        for d in device_list:
            sys.stdout.write("%s:\n" % d["Name"])
            sys.stdout.write("    inet: %s\n" % d["IPv4Address"])
            sys.stdout.write("    inet6: %s\n" % d["IPv6Address"])

    def __ip_get_address_type(self, address):
        # blind ipv4 check
        prot_type = LmiIpClient.IP_PROTOCOL_TYPE_IPV6
        if address.count(".") == 3:
            prot_type = LmiIpClient.IP_PROTOCOL_TYPE_IPV4
        return prot_type

    def __ip_get_ethernet_port_instance(self, device):
        return self._get_instances(LmiIpClient.LMI_ETHERNET_PORT_CLASS_NAME, Key = "DeviceID", Value = device)

    def __ip_get_ip_assignment_setting_data(self, device):
        (inst_list, rparams, errorstr) = self._get_instances(LmiIpClient.LMI_IP_ASSIGNMENT_SETTING_DATA_CLASS_NAME)
        if not inst_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = errorstr)
        inst_list = inst_list if type(inst_list) == list else [inst_list]
        rval_inst_list = filter(lambda x: x["Caption"] == "static-" + device, inst_list)
        if not rval_inst_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = "No IP setting for device %s found" % device)
        return LmiReturnValue(rval = rval_inst_list, rparams = rparams, errorstr = "")

    def __ip_remove_static_settings(self, device):
        (inst_list, rparams, errorstr) = self.__ip_get_ip_assignment_setting_data(device)
        if not inst_list:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        for inst in inst_list:
            (rval, rparams, errorstr) = self._delete_instance(inst.path)
            if rval != 0:
                return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        return LmiReturnValue(rval = True, rparams = rparams, errorstr = "")

    def __ip_create_static_setting(self, device, address, **kwargs):
        self.__ip_remove_static_settings(device)
        (device_inst, rparams, errorstr) = self.__ip_get_ethernet_port_instance(device)
        if not device_inst:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        if type(device_inst) is list:
            device_inst = device_inst[0]
        prot_type = self.__ip_get_address_type(address)
        args = {
            "Caption" : "static-" + device,
            "ProtocolIFType" : pywbem.Uint16(prot_type),
            "Port" : device_inst.path
        }
        if prot_type == LmiIpClient.IP_PROTOCOL_TYPE_IPV4:
            args["IPv4Address"] = LmiIpv4Addr(address).address
            if "SubnetMask" in kwargs:
                args["IPv4SubnetMask"] = kwargs["SubnetMask"]
            else:
                args["IPv4SubnetMask"] = LmiIpv4Addr(address).netmask
            if "DefaultGateway" in kwargs:
                args["IPv4DefaultGateway"] = kwargs["DefaultGateway"]
            if "DNSServers" in kwargs:
                args["IPv4DNSServers"] = kwargs["DNSServers"]
            if "SearchDomains" in kwargs:
                args["IPv4SearchDomains"] = kwargs["SearchDomains"]
        else:
            args["IPv6Address"] = address
            if "DefaultGateway" in kwargs:
                args["IPv6DefaultGateway"] = kwargs["DefaultGateway"]
            if "DNSServers" in kwargs:
                args["IPv6DNSServers"] = kwargs["DNSServers"]
            if "SearchDomains" in kwargs:
                args["IPv6SearchDomains"] = kwargs["SearchDomains"]
            if "PrefixLength" in kwargs:
                args["IPv6PrefixLength"] = pywbem.Uint8(kwargs["PrefixLength"])
        (configuration_inst, rparams, errorstr) = self._get_instances(LmiIpClient.IP_CONFIGURATION_SERVICE_CLASS_NAME)
        if not configuration_inst:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        if type(configuration_inst) is list:
            configuration_inst = configuration_inst[0]
        return self._call_method(configuration_inst, "CreateStaticSetting", **args)

    def __ip_apply_static_setting(self, device, setting_inst):
        (configuration_inst, rparams, errorstr) = self._get_instances(LmiIpClient.IP_CONFIGURATION_SERVICE_CLASS_NAME)
        if not configuration_inst:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        if type(configuration_inst) is list:
            configuration_inst = configuration_inst[0]
        (device_inst, rparams, errorstr) = self.__ip_get_ip_protocol_endpoint_instances(device)
        if type(device_inst) is list:
            device_inst = device_inst[0]
        if not device_inst:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        args = {
            "Configuration": setting_inst.path,
            "Endpoint" : device_inst.path
        }
        return self._call_method(configuration_inst, "ApplySettingToIPProtocolEndpoint", **args)

    def ip_list_devices(self, device = ""):
        (inst_list, rparams, errorstr) = self.__ip_get_ip_protocol_endpoint_instances(device)
        if not inst_list:
            return LmiReturnValue(rval = False, rparams = {}, errorstr = errorstr)
        self.__ip_print_device_info(inst_list)
        return LmiReturnValue(rval = True, rparams = rparams, errorstr = "")

    def ip_set_address(self, device, address, **kwargs):
        rval = self.__ip_create_static_setting(device, address, **kwargs)
        if not rval.rval:
            return rval
        setting_inst = self._cliconn.GetInstance(rval.rparams["SettingData"], LocalOnly = False)
        return self.__ip_apply_static_setting(device, setting_inst)

    def ip_del_address(self, device):
        return self.__ip_remove_static_settings(device)
