# Copyright (C) 2012-2013  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import abc
import new
import sys
import pywbem
import textwrap
import tempfile
import subprocess
from xml.dom.minicompat import NodeList
from lmi_client_base import LmiBaseClient
from lmi_client_base import LmiReturnValue

class LmiDeletedObjectError(Exception):
    """Raised, when there is an attempt to access properties
       on deleted _LmiInstance object"""

class LmiUnknownParameterError(Exception):
    """Raised, when there is a method call issued and unknown
       method parameter is provided"""

class LmiUnknownPropertyError(Exception):
    """Raised, when there is an attempt to create instance with
       unknown property provided"""

class LmiMethodCallError(Exception):
    """Raised, when an error occurs within method call"""

class LmiNoPagerError(Exception):
    """Raised, when tere is no default pager like less or more"""

def cast_lmi_to_cim(t, value):
    cast = {
        "sint8"  : lambda x: pywbem.Sint8(x),
        "uint8"  : lambda x: pywbem.Uint8(x),
        "sint16" : lambda x: pywbem.Sint16(x),
        "uint16" : lambda x: pywbem.Uint16(x),
        "sint32" : lambda x: pywbem.Sint32(x),
        "uint32" : lambda x: pywbem.Uint32(x),
        "sint64" : lambda x: pywbem.Sint64(x),
        "uint64" : lambda x: pywbem.Uint64(x),
        "reference": lambda x: x.path if isinstance(x, _LmiInstance) else x
    }
    cast_func = cast[t] if t.lower() in cast else lambda x: x
    if type(value) == list:
        return map(cast_func, value)
    elif type(value) == tuple:
        return tuple(map(cast_func, value))
    return cast_func(value)

class _LmiTextFormatter(object):
    def __init__(self, text):
        self._text = text

    def format(self, indent = 0, sub_indent = 0, width = 80, f = sys.stdout, **kwargs):
        if indent > width:
            return # NOTE: this is wrong!
        wrapper = textwrap.TextWrapper()
        wrapper.width = width - indent
        wrapper.subsequent_indent = " " * sub_indent
        for l in wrapper.wrap(self._text):
            f.write("%s%s\n" % (" " * indent, l))
        if "separator" in kwargs and kwargs["separator"]:
            f.write("\n")

class _LmiFormatter(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        stty_dimensions = os.popen("stty size 2> /dev/null", "r").read().split()
        self._width = int(stty_dimensions[1]) if stty_dimensions else 80

    @abc.abstractmethod
    def format(self, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        pass

    @staticmethod
    def __which(program):
        is_exe = lambda fpath: os.path.isfile(fpath) and os.access(fpath, os.X_OK)
        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None

    @staticmethod
    def __get_pager():
        if "PAGER" in os.environ:
            pager = os.environ["PAGER"].split()
            if os.path.exists(pager[0]) and os.access(pager[0], X_OK):
                return pager
        for p in ["less", "more"]:
            path = _LmiFormatter.__which(p)
            if not path:
                continue
            elif p == "less":
                return [path, "-S"]
            return [path]
        if LmiBaseClient._get_use_exceptions():
            raise LmiNoPagerError("No default pager found")
        return []

    def fancy_format(self, interactive):
        if interactive:
            tmpfile = tempfile.mkstemp()
            f = os.fdopen(tmpfile[0], "w")
            self.format(0, 0, self._width, f)
            f.close()
            subprocess_params = _LmiFormatter.__get_pager()
            if not subprocess_params:
                return
            subprocess_params.append(tmpfile[1])
            subprocess.call(subprocess_params)
            os.remove(tmpfile[1])
        else:
            self.format(0, 0, self._width)

class _LmiXmlFormatter(_LmiFormatter):
    def __init__(self, xml):
        super(self.__class__, self).__init__()
        self._xml = xml

    def __format_class_property_content(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        children = node.childNodes
        children_len = len(children)
        for (i, n) in enumerate(children):
            if isinstance(n, pywbem.cim_xml.QUALIFIER):
                self.__format_qualifier(n, indent, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__format_class_property(n, indent, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__format_class_property_array(n, indent, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.METHOD):
                self.__format_method(n, indent, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.VALUE):
                # NOTE: we skip default value for the class
                pass

    def __format_qualifier(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        is_array = isinstance(node.firstChild, pywbem.cim_xml.VALUE_ARRAY)
        val = "[qualifier] %s%s %s" % (node.getAttribute("TYPE"), " []" if is_array else "",
            node.getAttribute("NAME"))
        values = node.childNodes
        if values and isinstance(values[0], pywbem.cim_xml.VALUE_ARRAY):
            val += ": { " + ", ".join(["'" + v.firstChild.nodeValue + "'" for v in values[0].childNodes]) + " }\n"
        else:
            val += ": '%s'\n" % values[0].firstChild.nodeValue
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, f, separator = True)

    def __format_class_property(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        val = "[property] %s %s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, f, separator = True)
        self.__format_class_property_content(node, indent + 4, sub_indent, width, f)

    def __format_class_property_array(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        val = "[property array] %s [] %s" % (node.getAttribute("TYPE"), node.getAttribute("TYPE"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, f, separator = True)
        self.__format_class_property_content(node, indent + 4, sub_indent, width, f)

    def __format_class_property_reference(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        val = "[property ref] %s %s" % (node.getAttribute("REFERENCECLASS"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, f, separator = True)
        self.__format_class_property_content(node, indent + 4, sub_indent, width, f)

    def __format_instance_property(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        val = node.firstChild
        prop = "[property] %s %s%s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"),
            " = '%s'" % val.firstChild.nodeValue if val else "")
        _LmiTextFormatter(prop).format(indent, sub_indent + 4, width, f, separator = True)

    def __format_instance_property_array(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        prop = "[property array] %s [] %s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"))
        val = node.firstChild
        if val:
            prop += " = { " + ", ".join(["'" + v.firstChild.nodeValue + "'" for v in val.childNodes]) + " }"
        _LmiTextFormatter(prop).format(indent, sub_indent + 4, width, f, separator = True)

    def __format_instance_property_reference(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        val = "[property ref] %s %s" % (node.getAttribute("REFERENCECLASS"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, f, separator = True)

    def __format_method(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        qualifiers = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.QUALIFIER)]
        parameters = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER)]
        parameters_arr = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_ARRAY)]
        parameters_ref = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_REFERENCE)]
        parameters_ref_arr = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_REFARRAY)]
        has_args = parameters != [] or parameters_arr != [] or parameters_ref != [] or parameters_ref_arr != []
        val = "[method] %s %s%s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"), "(...)" if has_args else "()")
        _LmiTextFormatter(val).format(indent, sub_indent, width, f, separator = True)
        for q in qualifiers:
            self.__format_qualifier(q, indent + 4, sub_indent + 4, width, f)
        for p in parameters:
            val = "[param] %s %s" % (p.getAttribute("TYPE"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, f, separator = True)
        for p in parameters_arr:
            val = "[param array] %s (ref) %s" % (p.getAttribute("TYPE"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, f, separator = True)
        for p in parameters_ref:
            val = "[param ref] %s (ref) %s" % (p.getAttribute("REFERENCECLASS"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, f, separator = True)
        for p in parameters_ref_arr:
            val = "[param ref array] %s (ref) %s" % (p.getAttribute("REFERENCECLASS"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, f, separator = True)

    def __format_class(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        _LmiTextFormatter("Class: %s" % node.getAttribute("NAME")).format(indent, sub_indent, width, f)
        _LmiTextFormatter("SuperClass: %s" % node.getAttribute("SUPERCLASS")).format(indent + 4, sub_indent, width, f)
        for n in node.childNodes:
            if isinstance(n, pywbem.cim_xml.QUALIFIER):
                self.__format_qualifier(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__format_class_property(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__format_class_property_array(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_REFERENCE):
                self.__format_class_property_reference(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.METHOD):
                self.__format_method(n, indent + 4, sub_indent, width, f)

    def __format_instance(self, node, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        instance = node.firstChild.nextSibling
        _LmiTextFormatter("Instance of %s" % instance.getAttribute("CLASSNAME")).format(indent, sub_indent, width, f)
        for n in instance.childNodes:
            if isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__format_instance_property(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__format_instance_property_array(n, indent + 4, sub_indent, width, f)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_REFERENCE):
                self.__format_instance_property_reference(n, indent + 4, sub_indent, width, f)

    def format(self, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        if isinstance(self._xml, pywbem.cim_xml.CLASS):
            self.__format_class(self._xml, indent, sub_indent, width, f)
        elif isinstance(self._xml, pywbem.cim_xml.VALUE_NAMEDINSTANCE):
            self.__format_instance(self._xml, indent, sub_indent, width, f)

class _LmiMofFormatter(_LmiFormatter):
    def __init__(self, mof):
        super(self.__class__, self).__init__()
        self._mof = mof

    def format(self, indent = 0, sub_indent = 0, width = 80, f = sys.stdout):
        f.write(self._mof)

# Base class for classes, that are used for tab-completion.
# Used due to deprecated methods since python 2.2.
class BaseObject(object):
    # deprecated
    def __methods__(self):
        return []

    # deprecated
    def __members__(self):
        return []

    def __call__(self):
        pass

class _LmiNamespace(BaseObject):
    def __init__(self, client, name):
        self._client = client
        self._name = name

    def __getattr__(self, name):
        (klass, _, errorstr) = self._client._get_class(name, self._name, LocalOnly = False)
        if not klass:
            return None
        return _LmiClass(self._client, self, klass)

    def __repr__(self):
        return "%s(namespace='%s', ...)" % (self.__class__.__name__, self._name)

    def classes(self, filter_key = "", exact_match = False):
        (class_name_list, _, errorstr) = self._client._get_class_names(
            self._name, DeepInheritance = True)
        if not class_name_list:
            return []
        if filter_key:
            if not exact_match:
                filter_lambda = lambda n: n.lower().find(filter_key.lower()) >= 0
            else:
                filter_lambda = lambda n: n.lower() == filter_key.lower()
            class_name_list = filter(filter_lambda, class_name_list)
        return class_name_list

    def __transform_cim_to_lmi(self, cim_inst_list):
        lmi_inst_list = []
        for cim_inst in cim_inst_list:
            if isinstance(cim_inst, pywbem.cim_obj.CIMInstance):
                lmi_namespace = _LmiNamespace(self._client, cim_inst.path.namespace)
                (cim_class, _, errorstr) = self._client._get_class(cim_inst.classname,
                    cim_inst.path.namespace, LocalOnly = False)
                lmi_class = _LmiClass(self._client, lmi_namespace, cim_class)
                if not lmi_class:
                    return []
                lmi_inst_list.append(_LmiInstance(self._client, lmi_class, cim_inst))
            else:
                # NOTE: this is wrong, what should we do?
                return []
        return lmi_inst_list

    def print_classes(self, filter_key = "", exact_match = False):
        for c in self.classes(filter_key, exact_match):
            sys.stdout.write("%s\n" % c)

    def cql(self, query):
        (inst_list, _, errorstr) = self._client._exec_query(LmiBaseClient.QUERY_LANG_CQL, query, self._name)
        if not inst_list:
            return []
        return self.__transform_cim_to_lmi(inst_list)

    def wql(self, query):
        (inst_list, _, errorstr) = self._client._exec_query(LmiBaseClient.QUERY_LANG_WQL, query, self._name)
        if not inst_list:
            return []
        return self.__transform_cim_to_lmi(inst_list)

    @property
    def name(self):
        return self._name

class _LmiNamespaceRoot(_LmiNamespace):
    def __init__(self, client):
        super(self.__class__, self).__init__(client, "root")

    @property
    def cimv2(self):
        return _LmiNamespace(self._client, "root/cimv2")

    @property
    def interop(self):
        return _LmiNamespace(self._client, "root/interop")

    @property
    def PG_InterOp(self):
        return _LmiNamespace(self._client, "root/PG_InterOp")

    @property
    def PG_Internal(self):
        return _LmiNamespace(self._client, "root/PG_Internal")

    @property
    def namespaces(self):
        return ["cimv2", "interop", "PG_InterOp", "PG_Internal"]

    def print_namespaces(self):
        sys.stdout.write("cimv2\n")
        sys.stdout.write("interop\n")
        sys.stdout.write("PG_InterOp\n")
        sys.stdout.write("PG_Internal\n")

class _LmiClass(BaseObject):
    def __init__(self, client, namespace, klass):
        super(self.__class__, self).__init__()
        self._client = client
        self._namespace = namespace
        self._cim_class = klass

    def __repr__(self):
        return "%s(classname='%s', ...)" % (self.__class__.__name__, self.classname)

    def create_instance(self, properties = {}, qualifiers = {}, property_list = None):
        self_properties = self._cim_class.properties
        for (key, value) in properties.iteritems():
            if not key in self._cim_class.properties:
                if LmiBaseClient._get_use_exceptions():
                    raise LmiUnknownPropertyError
                return None
            t = self._cim_class.properties[key].type
            properties[key] = cast_lmi_to_cim(t, value)
        (cim_instance, rparams, errorstr) = self._client._create_instance(self.classname, self.namespace,
            properties, qualifiers, property_list)
        if not cim_instance:
            return None
        return _LmiInstance(self._client, self, cim_instance)

    def doc(self):
        _LmiXmlFormatter(self._cim_class.tocimxml()).fancy_format(self._client.interactive)

    def instance_names(self, **kwargs):
        (inst_name_list, _, errorstr) = self._client._get_instance_names(
            self._cim_class.classname, namespace = self._namespace.name, **kwargs)
        if not inst_name_list:
            return []
        result = []
        for i in inst_name_list:
            result.append(_LmiInstanceName(self._client, i))
        return result

    def first_instance_name(self, **kwargs):
        inst_name_list = self.instance_names(**kwargs)
        if not inst_name_list:
            return None
        return inst_name_list[0]

    def instances(self, **kwargs):
        (instance_list, _, errorstr) = self._client._get_instances(
            self._cim_class.classname, self._namespace.name, **kwargs)
        if not instance_list:
            return []
        return map(lambda inst: _LmiInstance(self._client, self, inst), instance_list)

    def first_instance(self, **kwargs):
        instance_list = self.instances(**kwargs)
        if not instance_list:
            return None
        return instance_list[0]

    def properties(self):
        return self._cim_class.properties.keys()

    def print_properties(self):
        for (name, prop) in self._cim_class.properties.iteritems():
            sys.stdout.write("%s\n" % name)

    def methods(self):
        return self._cim_class.methods.keys()

    def print_methods(self):
        for (name, method) in self._cim_class.methods.iteritems():
            sys.stdout.write("%s\n" % name)

    @property
    def classname(self):
        return self._cim_class.classname

    @property
    def namespace(self):
        return self._namespace.name

class _LmiInstanceName(BaseObject):
    def __init__(self, client, cim_instance_name):
        self._client = client
        self._cim_instance_name = cim_instance_name

    def __repr__(self):
        return "%s(...)" % self.__class__.__name__

    def to_instance(self):
        (cim_instance, _, errorstr) = self._client._get_instance(
            self._cim_instance_name, LocalOnly = False)
        if not cim_instance:
            return None
        lmi_namespace = _LmiNamespace(self._client, self._cim_instance_name.namespace)
        (cim_class, _, errorstr) = self._client._get_class(
            self._cim_instance_name.classname,
            self._cim_instance_name.namespace,
            LocalOnly = False)
        if not cim_class:
            return None
        lmi_class = _LmiClass(self._client, lmi_namespace, cim_class)
        if not lmi_class:
            return None
        return _LmiInstance(self._client, lmi_class, cim_instance)

    @property
    def path(self):
        return self._cim_instance_name

def possibly_deleted(fn):
    def wrapped(self = None, *args, **kwargs):
        if not self._deleted:
            return fn(self, *args, **kwargs)
        errorstr = "This instance has been deleted from a CIM broker"
        if LmiBaseClient._get_use_exceptions():
            raise LmiDeletedObjectError(errorstr)
        return LmiReturnValue(
            hostname = self._client.hostname,
            instance = self,
            use_display_hook = True,
            rval = None,
            rparams = {},
            errorstr = errorstr)
    return wrapped

class _LmiInstanceMethod(BaseObject):
    def __init__(self, client, lmi_instance, method):
        super(self.__class__, self).__init__()
        self._client = client
        self._lmi_instance = lmi_instance
        self._method = lmi_instance._lmi_class._cim_class.methods[method]

    def __transform_return_values(self, rparams):
        lmi_rparams = pywbem.NocaseDict()
        for (key, value) in rparams.iteritems():
            if isinstance(value, pywbem.cim_obj.CIMInstance):
                (lmi_class, _, errorstr) = self._client._get_class(value.classname,
                    value.namespace, LocalOnly = False)
                if not lmi_class:
                    return None
                lmi_rparams[key] = _LmiInstance(self._client, lmi_class, value)
            elif isinstance(value, pywbem.cim_obj.CIMInstanceName):
                lmi_rparams[key] = _LmiInstanceName(self._client, value)
            elif isinstance(value, pywbem.CIMInt):
                lmi_rparams[key] = int(value)
            elif isinstance(value, pywbem.CIMFloat):
                lmi_rparams[key] = float(value)
            elif type(value) == list:
                result = []
                for m in value:
                    if isinstance(m, pywbem.cim_obj.CIMInstance):
                        (lmi_class, _, errorstr) = self._client._get_class(m.classname,
                            m.namespace, LocalOnly = False)
                        if not lmi_class:
                            return None
                        lmi_rparams[key] = _LmiInstance(self._client, lmi_class, m)
                    elif isinstance(m, pywbem.cim_obj.CIMInstanceName):
                        result.append(_LmiInstanceName(self._client, m))
                lmi_rparams[key] = result
            else:
                lmi_rparams[key] = value
        return lmi_rparams

    def __call__(self, *args, **kwargs):
        for (param, value) in kwargs.iteritems():
            if param in self._method.parameters:
                t = self._method.parameters[param].type
                kwargs[param] = cast_lmi_to_cim(t, value)
            else:
                # NOTE: maybe we could check for pywbem type
                # and not to exit prematurely
                errorstr = "Unknown parameter '%s' supplied for method '%s'" % (param, self._method.name)
                if LmiBaseClient._get_use_exceptions():
                    raise LmiUnknownParameterError(errorstr)
                return LmiReturnValue(rval = -1, rparams = {},
                    errorstr = errorstr, hostname = self._client.hostname,
                    use_display_hook = True)
        (rval, call_rparams, call_errorstr) = self._client._call_method_raw(self._lmi_instance,
            self._method.name, **kwargs)
        if isinstance(rval, pywbem.CIMInt):
            rval = int(rval)
        elif isinstance(rval, pywbem.CIMFloat):
            rval = float(rval)
        if call_rparams:
            call_rparams = self.__transform_return_values(call_rparams)
            if not call_rparams:
                # NOTE: this is wrong! What should we do?
                if LmiBaseClient._get_use_exceptions():
                    raise LmiMethodCallError("Could not perform CIM -> LMI object transformation")
                return LmiReturnValue(rval = rval, rparams = {}, errorstr = errorstr)
        (new_inst, _, _) = self._client._get_instance(
            self._lmi_instance.path, LocalOnly = False)
        if not new_inst:
            # NOTE: this is wrong! What should we do?
            errorstr = "Could not update an LMI object after a method call"
            if LmiBaseClient._get_use_exceptions():
                raise LmiMethodCallError(errorstr)
            return LmiReturnValue(rval = rval, rparams = {}, errorstr = errorstr)
        self._lmi_instance._cim_instance = new_inst
        return LmiReturnValue(
            hostname = self._client.hostname,
            instance = self._lmi_instance,
            use_display_hook = True,
            rval = rval,
            rparams = call_rparams,
            errorstr = call_errorstr)

    @property
    def return_type(self):
        return self._method.return_type

    def tomof(self):
        _LmiMofFormatter(self._method.tomof()).fancy_format(self._client.interactive)

    def parameters(self):
        return self._method.parameters

    def print_parameters(self):
        for (param, value) in self._method.parameters.iteritems():
            sys.stdout.write("%s %s%s\n" % (value.type, param, "[]" if value.is_array else ""))

class _LmiInstance(BaseObject):
    def __init__(self, client, lmi_class, cim_instance):
        super(self.__class__, self).__init__()
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        self.__dict__["_deleted"] = False
        self.__dict__["_cim_instance"] = cim_instance
        self.__dict__["_client"] = client
        self.__dict__["_lmi_class"] = lmi_class

    @possibly_deleted
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        methods = self.methods()
        if methods and name in methods:
            return _LmiInstanceMethod(self._client, self, name)
        elif name in self._cim_instance:
            return self._cim_instance[name]
        raise AttributeError(name)

    @possibly_deleted
    def __setattr__(self, name, value):
        if name in self._cim_instance:
            self._cim_instance[name] = value
        else:
            self.__dict__[name] = value

    @possibly_deleted
    def __repr__(self):
        return "%s(classname='%s', ...)" % (self.__class__.__name__, self.classname)

    @property
    @possibly_deleted
    def classname(self):
        return self._lmi_class.classname

    @property
    @possibly_deleted
    def namespace(self):
        return self._lmi_class.namespace

    @property
    @possibly_deleted
    def path(self):
        return self._cim_instance.path

    @possibly_deleted
    def doc(self):
        _LmiXmlFormatter(self._cim_instance.tocimxml()).fancy_format(self._client.interactive)

    @possibly_deleted
    def tomof(self):
        _LmiMofFormatter(self._cim_instance.tomof()).fancy_format(self._client.interactive)

    @possibly_deleted
    def associator_names(self, AssocClass = None, ResultClass = None, Role = None,\
        ResultRole = None):
        result = []
        assoc_names_args = {}
        if AssocClass:
            assoc_names_args["AssocClass"] = AssocClass
        if ResultClass:
            assoc_names_args["ResultClass"] = ResultClass
        if Role:
            assoc_names_args["Role"] = Role
        if ResultRole:
            assoc_names_args["ResultRole"] = ResultRole
        assoc_names_list = self._client._get_associator_names(self._cim_instance,
            **assoc_names_args)
        for i in assoc_names_list:
            result.append(_LmiInstanceName(self._client, i))
        return result

    @possibly_deleted
    def first_associator_name(self, AssocClass = None, ResultClass = None, Role = None,
        ResultRole = None):
        result = self.associator_names(AssocClass = AssocClass, ResultClass = ResultClass,
            Role = Role, ResultRole = ResultRole)
        if not result:
            return []
        return result[0]

    @possibly_deleted
    def associators(self, AssocClass = None, ResultClass = None, Role = None,\
        ResultRole = None, Props = None):
        result = []
        assoc_args = {}
        if AssocClass:
            assoc_args["AssocClass"] = AssocClass
        if ResultClass:
            assoc_args["ResultClass"] = ResultClass
        if Role:
            assoc_args["Role"] = Role
        if ResultRole:
            assoc_args["ResultRole"] = ResultRole
        if Props:
            assoc_args["Props"] = Props
        associators_list = self._client._get_associators(self._cim_instance, **assoc_args)
        for i in associators_list:
            (cim_class, _, errorstr) = self._client._get_class(i.classname, i.path.namespace,
                LocalOnly = False)
            if not cim_class:
                return []
            lmi_namespace = _LmiNamespace(self._client, i.path.namespace)
            lmi_class = _LmiClass(self._client, lmi_namespace, cim_class)
            lmi_assoc_inst = _LmiInstance(self._client, lmi_class, i)
            result.append(lmi_assoc_inst)
        return result

    @possibly_deleted
    def first_associator(self, AssocClass = None, ResultClass = None, Role = None, \
        ResultRole = None, Props = None):
        result = self.associators(AssocClass = AssocClass, ResultClass = ResultClass,
            Role = Role, ResultRole = ResultRole, Props = Props)
        if not result:
            return None
        return result[0]

    @possibly_deleted
    def reference_names(self, ResultClass = None, Role = None):
        result = []
        reference_names_args = {}
        if ResultClass:
            reference_names_args["ResultClass"] = ResultClass
        if Role:
            reference_names_args["Role"] = Role
        reference_names_list = self._client._get_reference_names(
            self._cim_instance, **reference_names_args)
        for i in reference_names_list:
            result.append(_LmiInstanceName(self._client, i))
        return result

    @possibly_deleted
    def first_reference_name(self, ResultClass = None, Role = None):
        result = self.reference_names(ResultClass = ResultClass, Role = Role)
        if not result:
            return None
        return result[0]

    @possibly_deleted
    def references(self, ResultClass = None, Role = None, Props = None):
        result = []
        references_args = {}
        if ResultClass:
            references_args["ResultClass"] = ResultClass
        if Role:
            references_args["Role"] = Role
        if Props:
            references_args["Props"] = Props
        references_list = self._client._get_references(self._cim_instance, **references_args)
        for i in references_list:
            (cim_class, _, errorstr) = self._client._get_class(i.classname, i.path.namespace,
                LocalOnly = False)
            if not cim_class:
                return []
            lmi_namespace = _LmiNamespace(self._client, i.path.namespace)
            lmi_class = _LmiClass(self._client, lmi_namespace, cim_class)
            lmi_references_inst = _LmiInstance(self._client, lmi_class, i)
            result.append(lmi_references_inst)
        return result

    @possibly_deleted
    def first_reference(self, ResultClass = None, Role = None, Props = None):
        result = self.references(ResultClass = ResultClass, Role = Role, Props = Props)
        if not result:
            return None
        return result[0]

    @possibly_deleted
    def properties(self):
        return self._cim_instance.properties.keys()

    @possibly_deleted
    def print_properties(self):
        for (name, prop) in self._cim_instance.properties.iteritems():
            sys.stdout.write("%s\n" % name)

    @possibly_deleted
    def methods(self):
        return self._lmi_class.methods()

    @possibly_deleted
    def print_methods(self):
        self._lmi_class.print_methods()

    @possibly_deleted
    def push(self):
        return_value = self._client._modify_instance(self._cim_instance)
        return LmiReturnValue(
            hostname = self._client.hostname,
            instance = self._cim_instance,
            use_display_hook = True,
            rval = return_value.rval,
            rparams = return_value.rparams,
            errorstr = return_value.errorstr)

    @possibly_deleted
    def delete(self):
        rval = self._client._delete_instance(self.path)
        self._deleted = rval.rval == 0

class _LmiShellClient(LmiBaseClient):
    def __init__(self, hostname, username = "", password = "", interactive = False):
        super(self.__class__, self).__init__(hostname, username, password)
        self._interactive = interactive

    @property
    def interactive(self):
        return self._interactive

class LmiConnection(object):
    def __init__(self, hostname, username = "", password = "", interactive = False):
        self._shell_client = _LmiShellClient(hostname, username, password, interactive)

    @property
    def hostname(self):
        return self._shell_client._hostname

    @property
    def namespaces(self):
        return ["root"]

    def print_namespaces(self):
        sys.stdout.write("root\n")

    @property
    def root(self):
        return _LmiNamespaceRoot(self._shell_client)
